#  Copyright 2020 Oliver Cope
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any
from typing import Mapping
from typing import List
from typing import Tuple
import sys

from .parsing import BindParams, compile_bind_parameters
from . import exceptions


known_styles = {}


def get_param_style(conn: Any) -> str:

    conncls = conn.__class__
    try:
        return known_styles[conncls]
    except KeyError:
        modname = conncls.__module__
        while modname:
            try:
                style = sys.modules[modname].paramstyle
                known_styles[conncls] = style
                return style
            except AttributeError:
                if "." in modname:
                    modname = modname.rsplit(".", 1)[0]
                else:
                    raise TypeError(
                        f"Can't find paramstyle for connection {conn!r}"
                    )


class Query:

    name: str
    metadata: Mapping
    sql: str
    source: str

    def __init__(self, name, metadata, statements, source):
        self.name = name
        self.metadata = metadata
        self.statements = statements
        self.source = source
        self._conn = None

    def prepare(self, paramstyle, kw: Mapping) -> List[Tuple[str, BindParams]]:
        return [
            compile_bind_parameters(paramstyle, s, kw) for s in self.statements
        ]

    def bind(self, conn) -> "Query":
        """
        Return a copy of the query bound to a database connection
        """
        cls = self.__class__
        bound = cls.__new__(cls)
        bound.__dict__ = self.__dict__.copy()
        bound._conn = conn
        return bound

    def __call__(self, conn=None, *, debug=False, **kw):
        if conn is None:
            conn = self._conn
            if conn is None:
                raise TypeError(
                    "Query must be called with a connection argument"
                )
        rt = self.metadata["result"]

        paramstyle = get_param_style(conn)
        cursor = conn.cursor()
        affected = None

        for sqltext, bind_params in self.prepare(paramstyle, kw):
            if debug:
                import textwrap

                print(
                    f"Executing \n{textwrap.indent(sqltext, '    ')} with {bind_params!r}",
                    file=sys.stderr,
                )
            affected = cursor.execute(sqltext, bind_params)

        if rt == "one":
            return cursor.fetchone()

        if rt == "many":
            return iter(cursor.fetchone, None)

        if rt == "exactlyone":
            row = cursor.fetchone()
            if row is None:
                raise exceptions.NoResultFound()
            if cursor.fetchone() is not None:
                raise exceptions.MultipleResultsFound()
            return row

        if rt == "scalar":
            result = cursor.fetchone()
            if result is None:
                raise exceptions.NoResultFound()
            if isinstance(result, dict):
                return result[next(result)]
            return result[0]

        if rt == "column":
            return (row[0] for row in iter(cursor.fetchone, None))

        if rt == "affected":
            return affected

        if rt == "cursor":
            return cursor

        raise ValueError(f"Unsupported result type: {rt}")



#  Copyright 2020 Oliver Cope
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re
from itertools import count
from typing import List
from typing import Mapping
from typing import Union
from typing import Tuple

BindParams = Union[Tuple, Mapping]

DEFAULT_RESULT_TYPE = "cursor"

result_types = [
    ("one", ["one", "1"]),
    ("many", ["many", "*"]),
    ("affected", ["affected", "n"]),
    ("exactlyone", ["exactly-one", "=1"]),
    ("cursor", ["cursor", "raw"]),
    ("scalar", ["scalar"]),
    ("column", ["column"]),
]

result_type_pattern = rf"""
    :(?:
        {"|".join(
            f"(?P<{name}>{'|'.join(re.escape(t) for t in toks)})"
            for name, toks in result_types
        )}
    )
"""
name_pattern = re.compile(
    rf":name\s+(?P<name>\w+)(?:\s+{result_type_pattern})?$", re.X
)
result_pattern = re.compile(rf":result\s+{result_type_pattern}$", re.X)
param_pattern = re.compile(
    r"""
    # Don't match if preceded by backslash (an escape) or ':' (an SQL cast,
    # eg '::INT')
    (?<![:\\])

    # Optional parameter type
    (?:
        :(?P<param_type>
            value|v|value\*|v\*|tuple|t|tuple\*|t\*|identifier|i|raw|r
        )
    )?

    # An identifier
    :(?P<param>[a-zA-Z_]\w*)

    # followed by a non-word char, or end of string
    (?=\W|$)
    """,
    re.X,
)


def parse_comment_metadata(s: str) -> Mapping:

    lines = (re.sub(r"^\s*--", "", line).strip() for line in s.split("\n"))
    patterns = [name_pattern, result_pattern]
    result = {}
    for l in lines:
        for p in patterns:
            mo = p.match(l)
            if mo:
                for k, v in mo.groupdict().items():
                    if v is not None:
                        result[k] = v

    if result:
        return {
            "name": result.get("name", None),
            "result": next(
                (
                    rt
                    for rt in [name for name, _ in result_types]
                    if rt in result
                ),
                DEFAULT_RESULT_TYPE,
            ),
        }

    return None


def quote_ident_ansi(s):
    s = str(s)
    if "\x00" in s:
        raise ValueError(
            "Quoted identifiers can contain any character, "
            "except the character with code zero"
        )
    return f'''"{s.replace('"', '""')}"'''


def compile_bind_parameters(
    target_style: str, sql: str, bind_parameters: Mapping
) -> BindParams:
    """
    :param target_style: A DBAPI paramstyle value (eg 'qmark', 'format', etc)
    :param sql: An SQL str
    :bind_parameters: A dict of bind parameters for the query

    :return: tuple of `(sql, bind_parameters)`. ``sql`` will be rewritten with
             the target paramstyle; ``bind_parameters`` will be a tuple or
             dict as required.
    :raises: TypeError, if the bind_parameters do not match the query
    """

    positional, param_gen = {
        "qmark": (True, lambda name: "?"),
        "numeric": (True, lambda name, c=count(1): f":{next(c)}"),
        "format": (True, lambda name: "%s"),
        "pyformat": (False, lambda name: f"%({name})s"),
        "named": (False, lambda name: f":{name}"),
    }[target_style]

    if not bind_parameters:
        return (sql, (tuple() if positional else {}))

    if target_style[-6:] == "format":
        sql = sql.replace("%", "%%")

    positional_params = []
    named_params = {}

    def replace_placeholder(match):
        def make_sequence_placeholder(items: List, _c=count(1)) -> str:
            names = [f"_{next(_c)}" for _ in range(len(items))]
            if positional:
                positional_params.extend(items)
            else:
                named_params.update(zip(names, items))
            return ", ".join(param_gen(n) for n in names)

        group = match.groupdict().get
        pt = group("param_type")
        p = group("param")
        if pt in {None, "value", "v"}:
            if positional:
                positional_params.append(bind_parameters[p])
            else:
                named_params[p] = bind_parameters[p]
            return param_gen(p)
        elif pt in {"raw", "r"}:
            return str(bind_parameters.get(p))
        elif pt in {"identifier", "i"}:
            return quote_ident_ansi(bind_parameters.get(p))
        elif pt in {"value*", "v*", "tuple", "t"}:
            placeholder = make_sequence_placeholder(list(bind_parameters[p]))
            if pt in {"tuple", "t"}:
                return f"({placeholder})"
            return placeholder
        elif pt in {"tuple*", "t*"}:
            return ", ".join(
                f"({make_sequence_placeholder(list(items))})"
                for items in bind_parameters[p]
            )
        else:
            raise ValueError(f"Unsupported param_type {pt}")

    transformed_sql = param_pattern.sub(replace_placeholder, sql)
    if positional:
        return transformed_sql, tuple(positional_params)
    return transformed_sql, named_params



#  Copyright 2020 Oliver Cope
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import sqlparse
import pathlib
from itertools import chain
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import Mapping
from typing import Optional
from typing import Tuple
from typing import Union

from .query import Query
from .exceptions import InvalidStatement
from .parsing import parse_comment_metadata
from .parsing import DEFAULT_RESULT_TYPE

SQL_FILE_GLOB = "**/*.sql"


class Module:

    reloader: Optional[Callable] = None
    _conn = None
    queries: Dict = {}

    def __init__(self):
        self.queries = {}

    def __getattr__(self, name):
        if self.reloader:
            self.reloader()
        try:
            return self.queries[name]
        except KeyError:
            raise AttributeError(name)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        return

    def set_reloader(self, fn):
        self.reloader = fn

    def clear(self):
        self.queries.clear()

    def add_query(self, name, query):
        if name in self.queries:
            raise InvalidStatement(
                f"Can't add query {name!r} in {query.source}: "
                f"{self!r} already has an attribute named {name!r} "
                f"(loaded from {getattr(self, name).source})"
            )
        if self._conn:
            query = query.bind(self._conn)
        self.queries[name] = query

    def add_queries(
        self, qs: Union[Iterable[Tuple[str, Query]], Mapping[str, Query]]
    ):
        if isinstance(qs, Mapping):
            qs = qs.items()
        for name, query in qs:
            self.add_query(name, query)

    def load_dir(self, path: Union[str, pathlib.Path]):
        for p in pathlib.Path(path).glob(SQL_FILE_GLOB):
            self.load_file(p)

        return self

    def load_file(self, path: pathlib.Path):
        queries = list(load_queries(path))
        for query in queries:
            self.add_query(query.name, query)

    def execute(self, conn, sql=None, result="many", **kw):
        if sql is None:
            conn, sql = self._conn, conn
        query = Query(
            name=None,
            metadata={"result": result},
            statements=[sql],
            source="<string>",
        )
        return query(conn, **kw)

    def one(self, *args, **kwargs):
        return self.execute(result="one", *args, **kwargs)

    def many(self, *args, **kwargs):
        return self.execute(result="many", *args, **kwargs)

    def scalar(self, *args, **kwargs):
        return self.execute(result="scalar", *args, **kwargs)

    def affected(self, *args, **kwargs):
        return self.execute(result="affected", *args, **kwargs)

    def column(self, *args, **kwargs):
        return self.execute(result="column", *args, **kwargs)

    def cursor(self, *args, **kwargs):
        return self.execute(result="cursor", *args, **kwargs)

    def bind(self, conn) -> "Module":
        """
        Return a copy of the module bound to a database connection
        """
        cls = self.__class__
        bound = cls.__new__(cls)
        bound.__dict__ = {
            "_conn": conn,
            "reloader": self.reloader,
            "queries": {
                name: q.bind(conn) for name, q in self.queries.items()
            },
        }
        return bound

    def transaction(self, conn) -> "Transaction":
        return Transaction(self, conn)


class Transaction:
    def __init__(self, module, conn):
        self.conn = conn
        self.module = module.bind(conn)

    def __enter__(self):
        return self.module

    def __exit__(self, type, value, traceback):
        if type:
            self.conn.rollback()
        else:
            self.conn.commit()

    def commit(self):
        self.conn.commit()

    def rollback(self):
        self.conn.rollback()


def module(path, auto_reload=False):
    path = pathlib.Path(path)
    module = Module()
    module.load_dir(path)
    if auto_reload:

        def get_all_mtimes(path):
            return chain(
                [path.stat().st_mtime],
                (p.stat().st_mtime for p in path.glob(SQL_FILE_GLOB)),
            )

        module_mtime = [max(get_all_mtimes(path))]

        def reload_module():
            if any(m > module_mtime[0] for m in get_all_mtimes(path)):
                module_mtime[:] = [max(get_all_mtimes(path))]
                module.clear()
                module.load_dir(path)

        module.set_reloader(reload_module)
    return module


def _break_up_statements(
    path, sql
) -> Iterable[Tuple[Mapping, sqlparse.sql.Statement]]:
    statements = (s for s in sqlparse.parse(sql) if str(s).strip())
    current = ({}, [])
    for statement in statements:
        metadata = {}
        for token in statement.tokens[:]:
            if isinstance(token, sqlparse.sql.Comment):
                metadata = parse_comment_metadata(str(token))
                statement.tokens.remove(token)
                break

        if metadata:
            if current != ({}, []):
                yield current
            current = metadata, [str(statement)]

        else:
            current[1].append(str(statement))

    yield current


def load_queries(path: pathlib.Path) -> Iterable[Query]:
    with path.open("r", encoding="UTF-8") as f:
        sql = f.read()
        for ix, (metadata, statements) in enumerate(
            _break_up_statements(path, sql)
        ):
            if "result" not in metadata:
                metadata["result"] = DEFAULT_RESULT_TYPE
            if "name" not in metadata:
                if ix == 0:
                    metadata["name"] = path.stem
                else:
                    raise InvalidStatement(
                        f"{path!s}: no name specified (eg `-- :name my_query_name`)"
                    )
            if "result" not in metadata:
                raise InvalidStatement(
                    f"{path!s}: no result type specified (eg `-- :result :many`)"
                )
            yield Query(
                metadata["name"], metadata, statements, source=str(path)
            )
